/*
 * mpu6050.cpp
 *
 *  Created on: 19 mrt. 2015
 *      Author: steven
 *
 * The contents of this file have been heavily influenced by Jeff Rowbergs's work in the I2Cdev library collection.
 */

#include <cstring>
#include <cmath>
#include <iostream>

#include <unistd.h>

#include "mraa.hpp"
#include "mpu6050.hpp"

using namespace mraa;
using namespace std;


namespace df {

	// I2C address of the MPU6050
	const uint8_t mpu_addr = 0x68;

	/**
	 * The constructor does not initialize the IMU. Call init() to initialize the MPU6050
	 **/
	mpu6050::mpu6050()
		: m_i2c(nullptr), m_id(0), m_accel_resolution(1.0f), m_gyro_resolution(1.0f)
	{
		m_accel_bias.x = 0.0f;
		m_accel_bias.y = 0.0f;
		m_accel_bias.z = 0.0f;

		m_gyro_bias.x = 0.0f;
		m_gyro_bias.y = 0.0f;
		m_gyro_bias.z = 0.0f;
	}

	/**
	 * The destructor puts the MPU 6050 to sleep and closes the I2C bus.
	 **/
	mpu6050::~mpu6050()
	{
		if (m_i2c)
		{
			// Put the MPU6050 to sleep (low-power mode) by setting bit 7 of register 0x6B
			m_i2c->address(mpu_addr);
			m_i2c->writeReg(0x6B, 0x40);

			delete m_i2c;
		}
	}

	/**
	 * Calibrate the IMU, which means figuring out the bias
	 *
	 * @return 0 if the IMU was successfully calibrated
	 */
	int mpu6050::calibrate()
	{
		if (!m_i2c)
		{
			// The I2C pins are connected I2C bus 6 on the DF-Robot board.
			m_i2c = new I2c(6);
		}

		// Reset internal bias values
		m_accel_bias.x = m_accel_bias.y = m_accel_bias.z = 0;
		m_gyro_bias.x = m_gyro_bias.y = m_gyro_bias.z = 0;

		// Rest the device, such that all registers are cleared.
		if (!reset())
			return 1;

		// Use gyro X as the clock reference as advised in the documentation
		// The last 3 bits define the clock reference.
		// 0x01 = PLL with X axis gyroscope reference
		// 0x02 = PLL with Y axis gyroscope reference
		// 0x03 = PLL with Z axis gyroscope reference
		if (m_i2c->writeReg(0x6B, 0x01) != MRAA_SUCCESS)
			return 2;
		if (m_i2c->writeReg(0x6C, 0x00) != MRAA_SUCCESS)
			return 3;
		usleep(10000); // Why is this done???

		// Disable all interrupts
		if (m_i2c->writeReg(0x38, 0x01) != MRAA_SUCCESS)
			return 4;
		// Disable fifo use
		if (m_i2c->writeReg(0x23, 0x00) != MRAA_SUCCESS)
			return 5;
		// Turn on internal clock use
		if (m_i2c->writeReg(0x6B, 0x00) != MRAA_SUCCESS)
			return 6;
		// Disable I2C master control
		if (m_i2c->writeReg(0x24, 0x00) != MRAA_SUCCESS)
			return 7;
		// Disable FIFO and master modes
		if (m_i2c->writeReg(0x6A, 0x00) != MRAA_SUCCESS)
			return 8;
		if (!reset_fifo())
			return 9;
		usleep(10000); // TODO: Wait until flag has been reset instead of generic wait

		// Turn of all the external syncs and configure the DLPF (Digital Low-Pass Filter)
		// ID   | Accel bandwidth | Accel delay | Gyro bandwith | Gyro delay
		// -----|-----------------|-------------|---------------|------------
		// 0x00 |           260Hz |         0ms |         256Hz |    0.98ms
		// 0x01 |           184Hz |       2.0ms |         188Hz |    1.9ms
		// 0x02 |            94Hz |       3.0ms |          98Hz |    2.8ms
		// 0x03 |            44Hz |       4.9ms |          42Hz |    4.8ms
		// 0x04 |            21Hz |       8.5ms |          20Hz |    8.3ms
		// 0x05 |            10Hz |      13.8ms |          10Hz |   13.4ms
		// 0x06 |             5Hz |      19.0ms |           5Hz |   18.6ms
		if (m_i2c->writeReg(0x1A, 0x01) != MRAA_SUCCESS)
			return 10;

		// Use a 1KHz sample rate.
		if (m_i2c->writeReg(0x19, 0x00) != MRAA_SUCCESS)
			return 11;

		// Configure the accelerometer. The full scale range is determined by bits 3 and 4.
		// 16G = 0x18
		//  8G = 0x10
		//  4G = 0x08
		//  2G = 0x00
		m_accel_resolution = 2.0f / 32768.0f;
		if (m_i2c->writeReg(0x1C, 0x00) != MRAA_SUCCESS)
			return 12;

		// Configure the gyro. The full scale range is determined by bits 3 and 4.
		// 2000 degrees/s = 0x18
		// 1000 degrees/s = 0x10
		//  500 degrees/s = 0x08
		//  250 degrees/s = 0x00
		m_gyro_resolution = (250.0f / 32768.0f) * (M_PI / 180.0f); // The gyro resolution is also used to convert from degrees to rads
		if (m_i2c->writeReg(0x1B, 0x00) != MRAA_SUCCESS)
			return 13;

		/////////////////
		//             //
		// Sample data //
		//             //
		/////////////////

		// Enable storage of gyro and accel values in the fifo
		if (m_i2c->writeReg(0x23, 0x78) != MRAA_SUCCESS)
			return 14;
		// Enable the fifo
		if (!fifo_enable(true))
			return 15;

		// Wait until the fifo contains enough samples
		uint16_t fcount = fifo_count();
		int samples = 80;
		while(fcount < samples * 12)
		{
			usleep(1000);
			fcount = fifo_count();
		}

		// Disable fifo
		if (!fifo_enable(false))
			return 16;
		// Disable storage of gyro and accel values in the fifo
		if (m_i2c->writeReg(0x23, 0x00) != MRAA_SUCCESS)
			return 17;

		// Compute the average measurements
		uint32_t ax = 0;
		uint32_t ay = 0;
		uint32_t az = 0;
		uint32_t gx = 0;
		uint32_t gy = 0;
		uint32_t gz = 0;
		uint8_t raw_data[6];
		for (int i=0; i < samples; ++i)
		{
			// Read data from the fifo
			if (m_i2c->readBytesReg(0x74, raw_data, 6) != 6)
				return 16;
			uint16_t axt = (int16_t) (((int16_t)raw_data[0] << 8) | raw_data[1]);
			uint16_t ayt = (int16_t) (((int16_t)raw_data[2] << 8) | raw_data[3]);
			uint16_t azt = (int16_t) (((int16_t)raw_data[4] << 8) | raw_data[5]);
			uint16_t gxt = (int16_t) (((int16_t)raw_data[6] << 8) | raw_data[7]);
			uint16_t gyt = (int16_t) (((int16_t)raw_data[8] << 8) | raw_data[9]);
			uint16_t gzt = (int16_t) (((int16_t)raw_data[10] << 8) | raw_data[11]);

			ax += (uint32_t)axt;
			ay += (uint32_t)ayt;
			az += (uint32_t)azt;
			gx += (uint32_t)gxt;
			gy += (uint32_t)gyt;
			gz += (uint32_t)gzt;
		}

		// Clear the fifo
		reset_fifo();

		ax /= samples;
		ay /= samples;
		az /= samples;
		gx /= samples;
		gy /= samples;
		gz /= samples;

		// Remove gravity from the z-axis
		uint32_t accel_1_g = 16384; // The number of samples for 1 G
		if (az > 0)
			az -= accel_1_g;
		else
			az += accel_1_g;


		m_gyro_bias.x = (float)gx * m_gyro_resolution;
		m_gyro_bias.y = (float)gy * m_gyro_resolution;
		m_gyro_bias.z = (float)gz * m_gyro_resolution;

		cout << "---" << endl;

		cout << ax << "  " << ay << "  " << az << "  " << endl;
		cout << gx << "  " << gy << "  " << gz << "  " << endl;


		cout << m_gyro_bias.x << "  " << m_gyro_bias.y << "  " << m_gyro_bias.z << "  " << endl;

		// TODO

		if (!reset())
			return 18;

		return 0;
	}

	/**
	 * Initial the MPU6050
	 *
	 * @return 0 if the initialization was successful
	 **/
	int mpu6050::init()
	{
		if (!m_i2c)
		{
			// The I2C pins are connected I2C bus 6 on the DF-Robot board.
			m_i2c = new I2c(6);
		}

		if (!set_address())
			return 1;

		// Wake up the MPU6050 by setting bit 7 of register 0x6B to 0.
		if (m_i2c->writeReg(0x6B, 0) != MRAA_SUCCESS)
			return 2;

		// Give the IMU 10ms to wake-up
		usleep(10000);

		// Read the ID of the MPU6050
		m_id = m_i2c->readReg(0x75);

		// Use gyro X as the clock reference as advised in the documentation
		// The last 3 bits define the clock reference.
		// 0x01 = PLL with X axis gyroscope reference
		// 0x02 = PLL with Y axis gyroscope reference
		// 0x03 = PLL with Z axis gyroscope reference
		if (m_i2c->writeReg(0x6B, 0x01) != MRAA_SUCCESS)
			return 3;

		// Turn of all the external syncs and configure the DLPF (Digital Low-Pass Filter)
		// ID   | Accel bandwidth | Accel delay | Gyro bandwith | Gyro delay
		// -----|-----------------|-------------|---------------|------------
		// 0x00 |           260Hz |         0ms |         256Hz |    0.98ms
		// 0x01 |           184Hz |       2.0ms |         188Hz |    1.9ms
		// 0x02 |            94Hz |       3.0ms |          98Hz |    2.8ms
		// 0x03 |            44Hz |       4.9ms |          42Hz |    4.8ms
		// 0x04 |            21Hz |       8.5ms |          20Hz |    8.3ms
		// 0x05 |            10Hz |      13.8ms |          10Hz |   13.4ms
		// 0x06 |             5Hz |      19.0ms |           5Hz |   18.6ms
		if (m_i2c->writeReg(0x1A, 0x03) != MRAA_SUCCESS)
			return 4;

		// Use a 200Hz sample rate. I.e. 1KHz / (div + 1), which in this case is 1000 / (4 + 1) = 200
		if (m_i2c->writeReg(0x19, 0x04) != MRAA_SUCCESS)
			return 5;

		// Configure the accelerometer. The full scale range is determined by bits 3 and 4.
		// 16G = 0x18
		//  8G = 0x10
		//  4G = 0x08
		//  2G = 0x00
		m_accel_resolution = 2.0f / 32768.0f;
		if (m_i2c->writeReg(0x1C, 0x00) != MRAA_SUCCESS)
			return 6;

		// Configure the gyro. The full scale range is determined by bits 3 and 4.
		// 2000 degrees/s = 0x18
		// 1000 degrees/s = 0x10
		//  500 degrees/s = 0x08
		//  250 degrees/s = 0x00
		m_gyro_resolution = (250.0f / 32768.0f) * (M_PI / 180.0f); // The gyro resolution is also used to convert from degrees to rads
		if (m_i2c->writeReg(0x1B, 0x00) != MRAA_SUCCESS)
			return 7;

		// Enable the data ready interrupt
		if (m_i2c->writeReg(0x38, 0x01) != MRAA_SUCCESS)
			return 8;

		// Set the last update time, which is need to compute the delta t used in the quaternion update
		clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &m_last_q_update_time);

		return 0;
	}

	/**
	 * Inverse square root
	 */
	float mpu6050::inv_sqrt(float x) const
	{
		// Take from the example code by Sebastian Madgwick
		float halfx = 0.5f * x;
		float y = x;
		long i = *(long*)&y;
		i = 0x5f3759df - (i>>1);
		y = *(float*)&i;
		y = y * (1.5f - (halfx * y * y));
		return y;
	}

	/**
	 * Get the bytes written in the FIFO buffer
	 */
	uint16_t mpu6050::fifo_count() const
	{
		if (m_i2c->address(mpu_addr) != MRAA_SUCCESS)
			return 0;

		uint16_t hi = m_i2c->readReg(0x72);
		uint16_t lo = m_i2c->readReg(0x73);
		return (hi << 8) | lo;
	}

	/*
	 * Enable or disable the use of the FIFO buffer
	 *
	 * @param enable True if the use of the FIFO
	 */
	bool mpu6050::fifo_enable(bool enable) const
	{
		uint8_t fi = m_i2c->readReg(0x6A);
		fi = enable ? fi | 0x40: fi & 0xBF;
		if (m_i2c->writeReg(0x6A, fi) != MRAA_SUCCESS)
			return false;
		return true;
	}

	/**
	 * Reset the FIFO buffer
	 */
	bool mpu6050::reset_fifo() const
	{
		if (m_i2c->address(mpu_addr) != MRAA_SUCCESS)
			return false;

		uint8_t fi = m_i2c->readReg(0x6A);
		if (m_i2c->writeReg(0x6A, fi | 0x04) != MRAA_SUCCESS)
			return false;
		return true;
	}

	/**
	 * Read a 16-bit unsigned integer from two registers
	 *
	 * @param regHi The register with the high byte
	 * @param regLo The register with the low byte
	 **/
	short mpu6050::read_short(int regHi, int regLo) const
	{
		if (!set_address())
			return 0;

		// The Intel Edison is little endian, so the order is regLo, regHi
		short lo = m_i2c->readReg(regLo);
		short hi = m_i2c->readReg(regHi);
		return (hi << 8) | lo;
	}

	/**
	 * Reset the MPU6050 and wait until it is ready again
	 */
	bool mpu6050::reset() const
	{
		if (!set_address())
			return false;
		if (m_i2c->writeReg(0x6B, 0x80) != MRAA_SUCCESS)
			return false;

		// Wait until the device has reset
		while ((m_i2c->readReg(0x6B) & 0x80) != 0x00)
			usleep(10);

		return true;
	}

	/**
	 * Set the i2C address
	 *
	 * @return True if the address was successfully written.
	 */
	bool mpu6050::set_address() const
	{
		if (m_i2c == nullptr)
			return false;
		return m_i2c->address(mpu_addr) == MRAA_SUCCESS;
	}

	/**
	 * Read bytes
	 *
	 * @param data Buffer where to read bytes into
	 * @param length Number of bytes to read
	 * @return Number of bytes read
	 */
	uint8_t mpu6050::read_fifo(uint8_t* data, size_t length) const
	{
		if (!set_address())
			return 0;

		return m_i2c->readBytesReg(0x74, data, length);
	}

	/**
	 * Get the interrupt status
	 */
	uint8_t mpu6050::interrupt_status() const
	{
		return m_i2c->readReg(0x3A);
	}

	/**
	 * Get the x,y,z acceleration values
	 */
	mpu6050::vector3d mpu6050::read_acceleration() const
	{
		vector3d result;
		if (!set_address())
		{
			result.x = result.y = result.z = -999;
			return result;
		}

		uint8_t raw_data[6];
		if (m_i2c->readBytesReg(0x3B, raw_data, 6) != 6)
		{
			result.x = result.y = result.z = 999;
			return result;
		}

		int16_t x = (int16_t)(((int16_t)raw_data[0] << 8) | raw_data[1]);
		int16_t y = (int16_t)(((int16_t)raw_data[2] << 8) | raw_data[3]);
		int16_t z = (int16_t)(((int16_t)raw_data[4] << 8) | raw_data[5]);

		result.x = (float)x * m_accel_resolution - m_accel_bias.x;
		result.y = (float)y * m_accel_resolution - m_accel_bias.y;
		result.z = (float)z * m_accel_resolution - m_accel_bias.z;

		return result;
	}

	/**
	 * Get the x,y,z gyro values
	 */
	mpu6050::vector3d mpu6050::read_gyro() const
	{
		vector3d result;
		if (!set_address())
		{
			result.x = result.y = result.z = -999;
			return result;
		}

		uint8_t raw_data[6];
		if (m_i2c->readBytesReg(0x43, raw_data, 6) != 6)
		{
			result.x = result.y = result.z = 999;
			return result;
		}

		int16_t x = (int16_t)(((int16_t)raw_data[0] << 8) | raw_data[1]);
		int16_t y = (int16_t)(((int16_t)raw_data[2] << 8) | raw_data[3]);
		int16_t z = (int16_t)(((int16_t)raw_data[4] << 8) | raw_data[5]);

		result.x = (float)x * m_gyro_resolution - m_gyro_bias.x;
		result.y = (float)y * m_gyro_resolution - m_gyro_bias.y;
		result.z = (float)z * m_gyro_resolution - m_gyro_bias.z;


		return result;
	}

	/**
	 * Execute a Madgwick quaternion update.
	 *
	 * @param q The quaternion which is updated
	 */
	void mpu6050::madgick_imu_update(vector3d g, vector3d a, float dt)
	{
		// Convert values, such that Madgwick's is used. Remove this later when everything works
		float q0 = m_q[0];
		float q1 = m_q[1];
		float q2 = m_q[2];
		float q3 = m_q[3];

		float gx = g.x;
		float gy = g.y;
		float gz = g.z;

		float ax = a.x;
		float ay = a.y;
		float az = a.z;

		// The code below taken from the example code by Sebastian Madgwick
		// (http://www.x-io.co.uk/category/open-source/)

		float recipNorm;
		float s0, s1, s2, s3;
		float qDot1, qDot2, qDot3, qDot4;
		float _2q0, _2q1, _2q2, _2q3, _4q0, _4q1, _4q2 ,_8q1, _8q2, q0q0, q1q1, q2q2, q3q3;

		// Rate of change of quaternion from gyroscope
		qDot1 = 0.5f * (-q1 * gx - q2 * gy - q3 * gz);
		qDot2 = 0.5f * (q0 * gx + q2 * gz - q3 * gy);
		qDot3 = 0.5f * (q0 * gy - q1 * gz + q3 * gx);
		qDot4 = 0.5f * (q0 * gz + q1 * gy - q2 * gx);

		// Compute feedback only if accelerometer measurement valid (avoids NaN in accelerometer normalisation)
		if(!((ax == 0.0f) && (ay == 0.0f) && (az == 0.0f))) {

			// Normalise accelerometer measurement
			recipNorm = inv_sqrt(ax * ax + ay * ay + az * az);
			ax *= recipNorm;
			ay *= recipNorm;
			az *= recipNorm;

			// Auxiliary variables to avoid repeated arithmetic
			_2q0 = 2.0f * q0;
			_2q1 = 2.0f * q1;
			_2q2 = 2.0f * q2;
			_2q3 = 2.0f * q3;
			_4q0 = 4.0f * q0;
			_4q1 = 4.0f * q1;
			_4q2 = 4.0f * q2;
			_8q1 = 8.0f * q1;
			_8q2 = 8.0f * q2;
			q0q0 = q0 * q0;
			q1q1 = q1 * q1;
			q2q2 = q2 * q2;
			q3q3 = q3 * q3;

			// Gradient decent algorithm corrective step
			s0 = _4q0 * q2q2 + _2q2 * ax + _4q0 * q1q1 - _2q1 * ay;
			s1 = _4q1 * q3q3 - _2q3 * ax + 4.0f * q0q0 * q1 - _2q0 * ay - _4q1 + _8q1 * q1q1 + _8q1 * q2q2 + _4q1 * az;
			s2 = 4.0f * q0q0 * q2 + _2q0 * ax + _4q2 * q3q3 - _2q3 * ay - _4q2 + _8q2 * q1q1 + _8q2 * q2q2 + _4q2 * az;
			s3 = 4.0f * q1q1 * q3 - _2q1 * ax + 4.0f * q2q2 * q3 - _2q2 * ay;
			recipNorm = inv_sqrt(s0 * s0 + s1 * s1 + s2 * s2 + s3 * s3); // normalise step magnitude
			s0 *= recipNorm;
			s1 *= recipNorm;
			s2 *= recipNorm;
			s3 *= recipNorm;

			// Apply feedback step
			qDot1 -= beta * s0;
			qDot2 -= beta * s1;
			qDot3 -= beta * s2;
			qDot4 -= beta * s3;
		}

		// Integrate rate of change of quaternion to yield quaternion
		q0 += qDot1 * dt;
		q1 += qDot2 * dt;
		q2 += qDot3 * dt;
		q3 += qDot4 * dt;

		// Normalise quaternion
		recipNorm = inv_sqrt(q0 * q0 + q1 * q1 + q2 * q2 + q3 * q3);
		q0 *= recipNorm;
		q1 *= recipNorm;
		q2 *= recipNorm;
		q3 *= recipNorm;

		m_q[0] = q0;
		m_q[1] = q1;
		m_q[2] = q2;
		m_q[3] = q3;
	}

	/**
	 * ???
	 */
	mpu6050::orientation3d mpu6050::update()
	{
		vector3d a = read_acceleration();
		vector3d g = read_gyro();

		timespec now;
		clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &now);
		long dt_nanosecond = now.tv_nsec - m_last_q_update_time.tv_nsec;
		float dt = (float)(dt_nanosecond / 100) / 10000000.0f;
		m_last_q_update_time = now;

		cout << "dt=" << dt << endl;

		madgick_imu_update(g, a, 0.1);

		orientation3d result;
		result.yaw = atan2(2.0f * (m_q[1] * m_q[2] + m_q[0] * m_q[3]), m_q[0] * m_q[0] + m_q[1] * m_q[1] - m_q[2] * m_q[2] - m_q[3] * m_q[3]);
		result.pitch = -asin(2.0f * (m_q[1] * m_q[3] - m_q[0] * m_q[2]));
		result.roll = atan2(2.0f * (m_q[0] * m_q[1] + m_q[2] * m_q[3]), m_q[0] * m_q[0] - m_q[1] * m_q[1] - m_q[2] * m_q[2] + m_q[3] * m_q[3]);

		return result;
	}

} /* namespace df */
