/*
 * mpu6050.hpp
 *
 *  Created on: 19 mrt. 2015
 *      Author: steven
 */

#ifndef MPU6050_HPP_
#define MPU6050_HPP_

#include <cmath>
#include <time.h>

#include "mraa.hpp"

namespace df
{

/**
 * Get information from the MPU6050 IMU using only the information provided by the manual and open-source algorithms.
 */
class mpu6050
{
public:

	struct vector3d
	{
		float x;
		float y;
		float z;
	};

	struct orientation3d
	{
		float roll;
		float pitch;
		float yaw;
	};

private:

	// The I2C interface
	mraa::I2c* m_i2c;

	// ID of the IMU
	int m_id;

	// Vectors containing the bias, which need to be subtracted from the accelerometer and gyro readings
	vector3d m_accel_bias;
	vector3d m_gyro_bias;

	// The resolutions of the gyro and accelerometer
	float m_accel_resolution;
	float m_gyro_resolution;

	// Quaternion used by Madgwick algorithm
	float m_q[4];

	// Last time m_q was updated
	timespec m_last_q_update_time;

	const float gyro_meas_error = M_PI * (60.0f / 180.0f);
	const float beta = std::sqrt(3.0f / 4.0f) * gyro_meas_error;
	// TODO: I've seen beta's of 0.04 as well as 0.02 (with at least 100Hz sampling)

public:
	mpu6050();
	~mpu6050();
	mpu6050(const mpu6050&) = delete;
	mpu6050(mpu6050&&);

	mpu6050& operator=(const mpu6050&) = delete;
	mpu6050& operator=(mpu6050&&);

	/**
	 * Calibrate the IMU
	 */
	int calibrate();

	/**
	 * Wake-up the IMU initialize it for use.
	 */
	int init();

	/**
	 * Returns the temperature of the IMU
	 */
	float temperature() const
	{
		return float(read_short(0x41, 0x42)) / 340.0f + 36.53f;
	}

	/**
	 * Returns the ID of the IMU
	 */
	int who_am_i() const
	{
		return m_id & 0x7E;
	}

	/**
	 * Enable the use of the FIFO buffer
	 */
	bool fifo_enable(bool enable) const;

	/**
	 * Reset the FIFO buffer
	 */
	bool reset_fifo() const;

	/**
	 * Number of bytes in the FIFO buffer
	 */
	uint16_t fifo_count() const;

	/**
	 * Read length bytes into data from the FIFO buffer
	 *
	 * @return The number of bytes actually read.
	 */
	uint8_t read_fifo(uint8_t* data, size_t length) const;

	/**
	 * Interrupts state
	 */
	uint8_t interrupt_status() const;

	/**
	 * Get the x,y,z acceleration values in G
	 */
	vector3d read_acceleration() const;

	/**
	 * Get the x,y,z gyro values in rad/s
	 */
	vector3d read_gyro() const;

	/**
	 * Update the computed orientation and return it as a roll/pitch/yaw vector
	 */
	orientation3d update();

private:

	// Inverse sqrt
	float inv_sqrt(float x) const;

	// Madgwick quaternion update
	void madgick_imu_update(vector3d g, vector3d a, float dt);

	// Set the I2C address
	bool set_address() const;

	// Read an 16-bit unsigned integer from two registers
	short read_short(int regHi, int regLo) const;

	// Reset the IMU
	bool reset() const;
};


} /* namespace df */

#endif /* MPU6050_HPP_ */
