/*
 * mpu6050dmp.hpp
 *
 *  Created on: 6 apr. 2015
 *      Author: steven
 */

#ifndef MPU6050_DMP_HPP_
#define MPU6050_DMP_HPP_

#include "mraa.hpp"

namespace df
{

/**
 * Get information from the MPU6050 IMU using reverse engineered code for the DMP (Digital Motion Processor)
 */
class mpu6050_dmp
{
public:

	struct quaternion
	{
		float x;
		float y;
		float z;
		float w;
	};

	struct vector3d
	{
		float x;
		float y;
		float z;
	};

	struct orientation3d
	{
		float roll;
		float pitch;
		float yaw;
	};

private:

	// The I2C interface
	mraa::I2c* m_i2c;

	// ID of the IMU
	int m_id;

	const int packet_length = 42;

public:

	mpu6050_dmp();
	~mpu6050_dmp();
	mpu6050_dmp(const mpu6050_dmp&) = delete;
	mpu6050_dmp(mpu6050_dmp&&);

	mpu6050_dmp& operator=(const mpu6050_dmp&) = delete;
	mpu6050_dmp& operator=(mpu6050_dmp&&);

	/**
	 * Enable or disable the DMP. Once enabled, the FIFO buffer starts filling up.
	 */
	bool dmp_enable(bool enabled) const;

	/**
	 * Reset the DMP
	 */
	bool dmp_reset() const;

	/**
	 * Get the number of items in the FIFO buffer
	 */
	unsigned short fifo_count() const;

	/**
	 * Wake-up the IMU initialize the DMP (Digital Motion Processor)
	 */
	int init();

	/**
	 * Get the interrupt status and reset the interrupt
	 */
	unsigned short interrupt_status() const;

	/**
	 * Return true if a measurement is available.
	 */
	bool measurement_available() const;

	/**
	 * Read length bytes from the FIFO
	 */
	uint8_t read_fifo(uint8_t* data, size_t length) const;

	/**
	 * Get the orientation as a quaternion. Call this when measurement_available() is true.
	 */
	quaternion read_quaternion() const;

	/**
	 * Get the orientation as roll, pitch and yaw of the IMU. Call this when measurement_available() is true.
	 */
	orientation3d read_roll_pitch_yaw() const;

	/**
	 * Reset the FIFO buffer
	 */
	bool reset_fifo() const;

	/**
	 * Returns the temperature of the IMU
	 */
	float temperature() const
	{
		return float(read_short(0x41, 0x42)) / 340.0f + 36.53f;
	}

	/**
	 * Returns the ID of the IMU
	 */
	int who_am_i() const
	{
		return m_id & 0x7E;
	}

private:

	bool i2c_write_reg(uint8_t reg, uint8_t data) const;

	// Enable or disable the use of the FIFO buffer
	bool fifo_enable(bool enabled) const;

	// Read a block of data
	bool read_memory_block(uint8_t *data, uint16_t data_size, uint8_t bank, uint8_t address) const;

	// Read an 16-bit unsigned integer from two registers
	short read_short(int regHi, int regLo) const;

	// Reset the IMU
	bool reset() const;

	// Reset the i2c master
	bool reset_i2c_master() const;

	// Enable or disable I2C master mode
	bool set_i2c_master_mode_enabled(bool enabled) const;

	// Set the I2C address
	bool set_address() const;

	// Select memory bank
	bool set_memory_bank(uint8_t bank, bool prefetch_enabled, bool user_bank) const;

	// Set address within memory bank
	bool set_memory_start_address(uint8_t address) const;

	// Set I2C slave address
	bool set_slave_address(uint8_t num, uint8_t address) const;

	// Write a block of configuration data
	bool write_configuration_block(const uint8_t *data, uint16_t data_size) const;

	// Write a DMP memory block
	bool write_memory_block(const uint8_t *data, uint16_t data_size, uint8_t bank, uint8_t address, bool verify) const;

};

} /* namespace df */

#endif /* MPU6050_DMP_HPP_ */
